----------------------------------------------------------------------------------
--                                 _             _
--                                | |_  ___ _ __(_)__ _
--                                | ' \/ -_) '_ \ / _` |
--                                |_||_\___| .__/_\__,_|
--                                         |_|
--
----------------------------------------------------------------------------------
--
-- Company: hepia
-- Author: Joachim Schmidt <joachim.schmidt@hesge.ch>
--
-- Module Name: scalp_scheduler.vhd
-- Target Device: SCALP xc7z015clg485-2
-- Tool version: 2019.1
-- Description: Generic Round Robin or Random Scheduler.
--
-- Last update: 2020-01-17
--
---------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--use ieee.std_logic_unsigned.all;
--use ieee.std_logic_arith.all;
--use ieee.std_logic_misc.all;

--library UNISIM;
--use UNISIM.VCOMPONENTS.all;

library work;
use work.scalp_misc.all;
-- use work.scalp_utility.all;

entity scalp_scheduler is

    generic (
        C_SCALP_STRATEGY   : string   := "RR";  -- RR or RAND
        C_SCALP_PORTS      : positive := 6;
        C_SCALP_OUTPUT_REG : boolean  := true);

    port (
        -- Sys Clock and Sys Reset
        SysClkxCI        : in  std_ulogic;
        SysRstxRNAI      : in  std_ulogic;
        -- Request Side
        ArbitratexSI     : in  std_ulogic;
        RequestVectorxDI : in  std_ulogic_vector((C_SCALP_PORTS - 1) downto 0);
        -- Grant Side
        ArbitratedxSO    : out std_ulogic;
        GrantVectorxDO   : out std_ulogic_vector((C_SCALP_PORTS - 1) downto 0);
        GrantIndexxDO    : out std_ulogic_vector((scalp_log2ceilnz(C_SCALP_PORTS) - 1) downto 0));

end entity scalp_scheduler;

architecture behavioral of scalp_scheduler is

begin  -- architecture behavioral
    -- Assert C_SCALP_STRATEGY for known strings
    assert ((C_SCALP_STRATEGY = "RR"))  -- or (C_SCALP_STRATEGY = "RAND"))
        report "Unknow scheduler strategy." severity failure;

    -- Round Robin Scheduler (with RR strategy)
    RRStrategyxG : if C_SCALP_STRATEGY = "RR" generate
        signal RequestLeftxD      : unsigned((C_SCALP_PORTS - 1) downto 0)          := (others => '0');
        signal SelectLeftxD       : unsigned((C_SCALP_PORTS - 1) downto 0)          := (others => '0');
        signal SelectRightxD      : unsigned((C_SCALP_PORTS - 1) downto 0)          := (others => '0');
        signal ChannelPointerEnxS : std_ulogic                                      := '0';
        signal ChannelPointerxDP  : std_ulogic_vector((C_SCALP_PORTS - 1) downto 0) := scalp_to_sulv(1, C_SCALP_PORTS);
        signal ChannelPointerxDN  : std_ulogic_vector((C_SCALP_PORTS - 1) downto 0) := (others => '0');
    begin

        GlobalAsyncStatementsxB : block is
        begin  -- block GlobalAsyncStatementsxB
            ChannelPointerEnxAS : ChannelPointerEnxS <= ArbitratexSI;
            RequestLeftxAS      : RequestLeftxD      <= (not ((unsigned(ChannelPointerxDP) - 1) or unsigned(ChannelPointerxDP)) and unsigned(RequestVectorxDI));
            SelectLeftxAS       : SelectLeftxD       <= (unsigned(not RequestLeftxD) + 1) and RequestLeftxD;
            SelectRightxAS      : SelectRightxD      <= (unsigned(not RequestVectorxDI) + 1) and unsigned(RequestVectorxDI);
            ChannelPointerxAS   : ChannelPointerxDN  <= std_ulogic_vector(scalp_ite((RequestLeftxD = (RequestLeftxD'range => '0')), SelectRightxD, SelectLeftxD));
        end block GlobalAsyncStatementsxB;

        SchedulerxB : block is
        begin  -- block SchedulerxB            
            OutRegxG : if not C_SCALP_OUTPUT_REG generate  -- Without Output Register
                AsyncStatementsxB : block is
                begin  -- block AsyncStatementsxB
                    ArbitratedxAS  : ArbitratedxSO  <= ArbitratexSI;
                    GrantVectorxAS : GrantVectorxDO <= ChannelPointerxDN;
                    GrantIndexxAS  : GrantIndexxDO  <= std_ulogic_vector(scalp_onehot2bin(ChannelPointerxDN));
                end block AsyncStatementsxB;

                UpdateRegxP : process (SysClkxCI, SysRstxRNAI) is
                begin  -- process UpdateRegxP
                    if SysRstxRNAI = '0' then
                        ChannelPointerxDP <= scalp_to_sulv(1, C_SCALP_PORTS);
                    elsif rising_edge(SysClkxCI) then
                        if ChannelPointerEnxS = '1' then
                            ChannelPointerxDP <= ChannelPointerxDN;
                        end if;
                    end if;
                end process UpdateRegxP;
            elsif C_SCALP_OUTPUT_REG generate  -- With Output Register
                signal ChannelPointerBinxDP : std_ulogic_vector((scalp_log2ceilnz(C_SCALP_PORTS) - 1) downto 0) := (others => '0');
            begin
                AsyncStatementsxB : block is
                begin  -- block AsyncStatementsxB
                    ArbitratedxAS  : ArbitratedxSO  <= ArbitratexSI when rising_edge(SysClkxCI);
                    GrantVectorxAS : GrantVectorxDO <= ChannelPointerxDP;
                    GrantIndexxAS  : GrantIndexxDO  <= ChannelPointerBinxDP;
                end block AsyncStatementsxB;

                UpdateRegxP : process (SysClkxCI, SysRstxRNAI) is
                begin  -- process UpdateRegxP
                    if SysRstxRNAI = '0' then
                        ChannelPointerxDP    <= scalp_to_sulv(1, C_SCALP_PORTS);
                        ChannelPointerBinxDP <= (others => '0');
                    elsif rising_edge(SysClkxCI) then
                        if ChannelPointerEnxS = '1' then
                            ChannelPointerxDP    <= ChannelPointerxDN;
                            ChannelPointerBinxDP <= std_ulogic_vector(scalp_onehot2bin(ChannelPointerxDN));
                        end if;
                    end if;
                end process UpdateRegxP;
            end generate OutRegxG;
        end block SchedulerxB;
    end generate RRStrategyxG;

end architecture behavioral;
